const { red, blue, green } = require('chalk');
const https = require('https');
const findPackageJSON = require('find-package-json');
const detectIndent = require('detect-indent');
const fs = require('fs');
const path = require('path');
const Stream = require('stream').Transform;
const glob = require('glob');
const {
  TEMPLATES,
  IGNORED_DIRECTORIES,
  IGNORED_FILES,
} = require('../constants');

function logError(message) {
  console.error(`❌ ${red(message)}`);
}

function logVerbose(message, ...rest) {
  if (global.verbose || process.env.DEBUG) {
    console.log(`${blue('verbose:')} ${message}`, ...rest);
  }
}

function logSuccess(message) {
  console.log(green(message));
}

function logInfo(message) {
  console.log(blue(message));
}

function logHelp(command) {
  const mainHelp = `
Usage:
  blazepack   [options]  <command>

Options:
  --version   Output the current version
  --port <p>  Specify port <p> for dev server
  --help      Provides help

Commands:
  start           Starts the dev server
  create          Create project from Template
  clone           Clone an existing codesandbox
  export          Export to codesandbox.io
  install         Install dependency
  i               Install dependency (alias for install)
  add             Install dependency (alias for install)
  uninstall       Remove dependency
  remove          Remove dependency (alias for uninstall)
`;
  if (!command) {
    console.log(mainHelp);
  } else {
    switch (command) {
      case 'create':
        console.log(`
Usage: blazepack create <name> [--template=<template>][--open]
        
Create project with name <name> from Template <template>.
If the --template option is not specified you will get the list of available templates to select from.
If the --open options is specified then dev server will be stated in the newly created project.

eg: blazepack create my-cra --template=react --open

List of available templates are:`);
        Object.keys(TEMPLATES).forEach((template) => {
          console.log(`  ${template}`);
        });
        break;
      case 'start':
        console.log(`
Usage: blazepack start [directory][--port=<port>]
        
Starts the dev server in the given directory.
If no directory is specified then the current working directory is used.
You can change the port on which the dev server runs using the --port option.
`);
        break;
      case 'export':
        console.log(`
Usage: blazepack export [--open]

Export your current project to codesandbox.io.
Opens the sandbox in a browser tab if the --open option is specified.
`);
        break;
      case 'clone':
        console.log(`
Usage: blazepack clone <sandbox-url|sandbox-id|embed-url>.

Clone the sandbox from the given sadbox url, id or embed-url.
`);
        break;
      case 'i':
      case 'install':
      case 'add':
        console.log(`\nUsage: blazepack ${command} <package>`);
        console.log(`
Quickly Install a new package (dependency).
It does not create node_modules so you are gonna save a lot of space.
The install, i, add commands are all similar.

eg: blazepack install redux

# or

blazepack i redux

# or

blazepack add redux
`);
        break;
      case 'uninstall':
      case 'remove':
        console.log(`\nUsage: blazepack ${command} <package>`);
        console.log(`
Quickly Remove an unused package (dependency).
The uninstall & remove command are both similar.

eg: blazepack uninstall redux

# or

blazepack remove redux
`);
        break;
      default:
        console.log(`Error: unknown command '${command}'`);
        console.log(mainHelp);
        break;
    }
  }
}

function getExtension(filename) {
  const fileParts = filename.split('.');

  return fileParts.length > 1 ? fileParts[fileParts.length - 1] : null;
}

function isImage(filename) {
  const ext = getExtension(filename);
  const imageExtensions = ['png', 'jpg', 'jpeg', 'gif', 'webp'];

  return imageExtensions.includes(ext);
}

function getSandboxFiles(id) {
  return new Promise((resolve, reject) => {
    https
      .get(`https://codesandbox.io/api/v1/sandboxes/${id}`, (response) => {
        if (response.statusCode == 200) {
          let data = '';
          response.on('data', (chunk) => {
            data += chunk;
          });

          response.on('end', () => {
            resolve(JSON.parse(data));
          });
        } else {
          reject(`failed to download sandbox for ${id}`);
        }
      })
      .on('error', (e) => {
        reject(e);
      });
  });
}

function downloadFile(url) {
  return new Promise((resolve, reject) => {
    https
      .get(url, (response) => {
        if (response.statusCode == 200) {
          let data = new Stream();
          response.on('data', (chunk) => {
            data.push(chunk);
          });

          response.on('end', () => {
            resolve(data.read());
          });
        } else {
          reject(`failed to download image ${url}`);
        }
      })
      .on('error', (e) => {
        reject(e);
      });
  });
}

async function createSandboxFiles(sandboxInfo, projectName) {
  /**
   * Object of all directories
   *
   * For eg:
   * {
   *    "a": {
   *      directory_shortid: "b", // components directory resides inside the directory with id `b`
   *      title: "components" // directory name
   *    },
   *    "b": {
   *      directory_shortid: null, // does not have parent directory
   *      title: "src"
   *    },
   * }
   *
   */
  // create an empty dir for our project
  const projectDir = projectName || sandboxInfo.title || sandboxInfo.id;
  if (projectDir)
    fs.mkdirSync(projectName || sandboxInfo.title || sandboxInfo.id);

  const directories = sandboxInfo.directories.reduce(
    (agg, directory) => ({
      ...agg,
      [directory.shortid]: {
        directory_shortid: directory.directory_shortid,
        title: directory.title,
      },
    }),
    {}
  );

  /**
   * Get directory name recursively
   */
  const getFolderName = (id, currentDir = '') => {
    if (!id || !directories[id]) {
      return currentDir;
    }

    if (directories[id].directory_shortid) {
      return getFolderName(
        directories[id].directory_shortid,
        `${directories[id].title}/${currentDir}`
      );
    }

    return `${directories[id].title}/${currentDir}`;
  };

  /**
   * Directories object with nested path
   *
   * {
   *    "a": "src/components/",
   *    "b": "src/",
   * }
   *
   */
  const directoriesWithPath = Object.keys(directories).reduce((agg, dir) => {
    return { ...agg, [dir]: getFolderName(dir) };
  }, {});

  /**
   * Project path with the sandbox name
   */
  const projectPath = path.join(process.cwd(), projectDir);

  /**
   * Create all directories, recursive: true; forces the directory creation if not present
   */
  Object.keys(directoriesWithPath).forEach((dir) => {
    fs.mkdirSync(`${projectPath}/${directoriesWithPath[dir]}`, {
      recursive: true,
    });
  });

  /**
   * Create all files, with the code.
   */
  for (const module of sandboxInfo.modules) {
    // if it is a url we need to download it from codesandbox locally
    if (/^https?:\/\/(.*)$/.test(module.code)) {
      try {
        module.code = await downloadFile(module.code);
      } catch (err) {
        logError(`${module.title}: ${err}`);
      }
    }

    if (module.directory_shortid) {
      fs.writeFileSync(
        `${projectPath}/${directoriesWithPath[module.directory_shortid]}${
          module.title
        }`,
        module.code
      );
    } else {
      fs.writeFileSync(`${projectPath}/${module.title}`, module.code);
    }
  }
}

function getPackageJSON() {
  const iterator = findPackageJSON();
  const nextPackageJSON = iterator.next();

  if (nextPackageJSON && nextPackageJSON.filename) {
    const packageJSONContent = fs.readFileSync(
      nextPackageJSON.filename,
      'utf-8'
    );

    const packageJSON = JSON.parse(packageJSONContent);

    return {
      file: nextPackageJSON.filename,
      indent: detectIndent(packageJSONContent).indent || 2,
      json: packageJSON,
    };
  }

  return null;
}

function getProjectRoot(directory) {
  const iterator = findPackageJSON(directory);
  const nextPackageJSON = iterator.next();

  if (nextPackageJSON && nextPackageJSON.filename) {
    return path.dirname(nextPackageJSON.filename);
  }

  return null;
}

function getPosixPath(filePath) {
  return filePath.split(path.sep).join(path.posix.sep);
}

function readAsDataUrlSync(filePath) {
  const filename = path.basename(filePath);
  const ext = getExtension(filename);
  const fileContent = fs.readFileSync(filePath, 'base64');

  return `data:image/${ext};base64,${fileContent}`;
}

function readSandboxFromFS(directory, exportFormat = false) {
  let sandboxFiles = {};
  const ignoredDirGlob = IGNORED_DIRECTORIES.map(
    (IGNORED_DIRECTORY) => `**/${IGNORED_DIRECTORY.source}/**`
  );
  const ignoredFileGlob = IGNORED_FILES.map(
    (IGNORED_FILE) => `**/${IGNORED_FILE.source}`
  );

  // get all files in the dir except ignored directories and files
  const filePaths = glob.sync(`${directory}/**`, {
    absolute: true,
    ignore: [...ignoredDirGlob, ...ignoredFileGlob],
    nodir: true,
  });

  logVerbose('glob found files:', filePaths);

  filePaths.forEach((filePath) => {
    const filename = path.basename(filePath);
    const relativePath = getPosixPath(path.relative(directory, filePath));
    let fileContent;

    if (isImage(filename)) {
      fileContent = readAsDataUrlSync(filePath);
    } else {
      fileContent = fs.readFileSync(filePath, 'utf-8');
    }

    const sandboxFilePath = `/${relativePath}`;

    // we use the export format in the export sandbox command
    sandboxFiles[sandboxFilePath] = exportFormat
      ? {
          content: fileContent,
        }
      : {
          code: fileContent,
          path: sandboxFilePath,
        };
  });

  return sandboxFiles;
}

function hasDependency(packageJSON, dependency) {
  const dependencies = Object.keys(packageJSON.dependencies || {});
  const devDependencies = Object.keys(packageJSON.devDependencies || {});

  return (
    dependencies.includes(dependency) || devDependencies.includes(dependency)
  );
}

function detectTemplate(directory) {
  const sandboxConfig = path.join(directory, 'sandbox.config.json');
  const packageJSON = path.join(directory, 'package.json');

  if (!fs.existsSync(packageJSON) && !fs.existsSync(sandboxConfig)) {
    throw 'Unknown project template!';
  }

  if (fs.existsSync(sandboxConfig)) {
    try {
      const sandboxConfigContent = JSON.parse(fs.readFileSync(sandboxConfig));

      return sandboxConfigContent.template;
    } catch {
      throw 'Invalid sandbox.config.json';
    }
  }

  if (fs.existsSync(packageJSON)) {
    try {
      const packageJSONContent = JSON.parse(fs.readFileSync(packageJSON));

      if (hasDependency(packageJSONContent, 'react-scripts')) {
        return 'react';
      }
      if (hasDependency(packageJSONContent, 'svelte')) {
        return 'svelte';
      }
      if (hasDependency(packageJSONContent, 'parcel-bundler')) {
        return 'parcel';
      }
      if (hasDependency(packageJSONContent, 'preact')) {
        return 'preact';
      }
      if (hasDependency(packageJSONContent, '@vue/cli-service')) {
        return 'vue';
      }
      if (hasDependency(packageJSONContent, '@angular/core')) {
        return 'angular';
      }
      if (hasDependency(packageJSONContent, 'reason-react')) {
        return 'reason-reason';
      }
      if (hasDependency(packageJSONContent, '@dojo/cli')) {
        return 'dojo';
      }
      if (hasDependency(packageJSONContent, 'cx-react')) {
        return 'cxjs';
      }
    } catch {
      throw 'Invalid package.json';
    }

    throw 'Unknown project template!';
  }
}

module.exports = {
  logError,
  logInfo,
  logSuccess,
  logHelp,
  logVerbose,
  getSandboxFiles,
  createSandboxFiles,
  isImage,
  getExtension,
  getPackageJSON,
  readSandboxFromFS,
  getPosixPath,
  detectTemplate,
  getProjectRoot,
};
